#' Plot PCA
#'
#' Plot multi-dimensional scaling plot using algorithm of BiocSingular::runPCA(). It is recommended this be done with
#' the log-methylation-ratio matrix generated by bsseq_to_log_methy_ratio().
#'
#' @param x the log-methylation-ratio matrix.
#' @param plot_dims the numeric vector of the two dimensions to be plotted.
#' @param labels the character vector of labels for data points. By default uses column names of x, set to NULL to plot
#'   points.
#' @param groups the character vector of groups the data points will be coloured by.
#'
#' @return ggplot object of the MDS plot.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' bss <- methy_to_bsseq(nmr)
#' lmr <- bsseq_to_log_methy_ratio(bss)
#' plot_pca(lmr)
#'
#' @importFrom BiocSingular runPCA
#'
#' @export
plot_pca <- function(x, plot_dims = c(1, 2), labels = colnames(x), groups = NULL) {
    if (!is.null(labels)) {
        assertthat::assert_that(ncol(x) == length(labels))
    }

    if (!is.null(groups)) {
        assertthat::assert_that(ncol(x) == length(groups))
    }

    pca_res <- BiocSingular::runPCA(t(x), rank = max(plot_dims))

    plot_data <- data.frame(
        dim1 = pca_res$x[, plot_dims[1]],
        dim2 = pca_res$x[, plot_dims[2]]
    )

    plot_data$labels <- labels
    plot_data$groups <- groups

    xlabel <- glue::glue("PCA Dim {plot_dims[1]}")
    ylabel <- glue::glue("PCA Dim {plot_dims[2]}")

    if (!is.null(groups)) {
        p <- ggplot2::ggplot(plot_data, ggplot2::aes_string(x = "dim1", y = "dim2", col = "groups"))
    } else {
        p <- ggplot2::ggplot(plot_data, ggplot2::aes_string(x = "dim1", y = "dim2"))
    }

    if (!is.null(labels)) {
        p <- p + ggplot2::geom_label(ggplot2::aes_string(label = "labels"))
    } else {
        p <- p + ggplot2::geom_point()

    }

    p +
        ggplot2::theme_bw() +
        ggplot2::xlab(xlabel) +
        ggplot2::ylab(ylabel) +
        ggplot2::scale_x_continuous(expand = c(.1, .1))
}
